from flask import Flask, render_template, request, send_from_directory, abort, jsonify, Response
from flask import session, redirect
from flaskext.markdown import Markdown
from wtforms import Form
import jinja2_highlight
import os
import traceback
import base64

from rocolib.library import getComponent, getComponentPaths
from rocolib.api.composables.graph.Joint import FingerJoint

from widgets import * # ButtonField defined here
from known_models import known_models


# Hack to make thin STLs behave nicer
MIN_THICKNESS = 0.05

# https://github.com/tlatsas/jinja2-highlight/blob/master/examples/flask/flask-example.py
jinja_options = dict(Flask.jinja_options)
jinja_options.setdefault('extensions', []).append('jinja2_highlight.HighlightExtension')

app = Flask(__name__)
Markdown(app)
app.secret_key = "ROCOSecretKey"

@app.route('/basic')
def setold(): 
    session['basic'] = True
    return redirect('/')

@app.route('/modern')
def setnew(): 
    session['basic'] = False
    return redirect('/')

# Relative to environment variable $REPO if exists, otherwise default
git_repo = os.environ.get("REPO", "https://git.uclalemur.com/nrobot/oldroco")
if git_repo[-4:] == ".git":
    git_repo = git_repo[:-4]
git_files = git_repo + "/-/tree/master/rocolib"

def getModels(tfn):
    return [k for k in known_models.keys() if tfn(known_models[k].get("t", 0))]

jinja_model_info = {
    'Wood Models': getModels(lambda x : x > 1),
    'Paper Models': getModels(lambda x : x <= 1),
}

known_types = {
    "svg": ("unfolding", "image/svg+xml"),
    "dxf": ("silhouette", "application/dxf"),
    "stl": ("stl", "application/sla"),
    "png": ("png", "image/png"),
}

def make(f, path, args, mkstl=True, mkdxf=True, mksvg=True, mkpng=True): 
    # Default thickness: paper unless we know otherwise
    t = known_models.get(path, {}).get("t", 0)

    # args is a dict of parameter/value pairs
    for p, v in args.items():
        try:
            # I guess we're only considering numeric parameters?
            # and it might cause problems when integer valued parameters are set as floats
            # TODO: we should probably figure out how to handle v based on p's parameter metadata 
            value = float(v)
            if value == int(value):
                value = int(value)
            f.setParameter(p, value)
        except KeyError:
            # 'thickness' is an argument to makeOutput
            # there are others, but we're not handling them yet
            # We don't distinguish between these argument vs. component parameters in the frontend
            # but we need to handle them appropriately
            if p == '$thickness' or p == '%24thickness':
                t = float(v)
        except ValueError as e:
            # parameter name not found in Component object, ignore
            print("*** ERROR ***", repr(e))
            traceback.print_exc()
            pass

    t = max(t, MIN_THICKNESS)
    j = None
    if t > 1:
        j = FingerJoint(thickness=t)

    try:
        ss = f.makeOutput(None, default=False,
            unfolding=mksvg, silhouette=mkdxf, stl=mkstl, png=mkpng,
            thickness = t, joint = j)
    except Exception as e:
        print("*** ERROR ***", repr(e))
        traceback.print_exc()
        return repr(e)
    return ss["graph"]

@app.route("/")
@app.route("/home")
def home():
    if session.get("basic", False):
        string = "<html>"
        string += '<head><link rel="shortcut icon" href="static/favicon.ico"></head>'
        string += "<body>"
        for m in known_models.keys():
            string += '<a href="' + m + '">' + m + "</a><br>"
            string += "</body></html>"
        return string

    return render_template("home.html", model_info=jinja_model_info)

# Ruta con condicionales y parametros
@app.route("/about")
def about():
    return render_template("about.html")
    #numero = randint(0, 19)
    #return render_template("index2.html", numero=numero)

@app.route("/tutorial")
def _tutorial():
    return render_template("tutorial.html")

@app.route("/python_tutorial")
def _python():
    return render_template("python_tutorial.html", git_repo=git_repo, git_files=git_files)

@app.route('/favicon.ico')
def favicon():
    return send_from_directory(os.path.join(app.root_path, 'static'),
                               'favicon.ico', mimetype='image/vnd.microsoft.icon')

# --- Component Functionality --- #

@app.route('/<path:path>', methods=['GET', 'POST'])
def catchall(path): 
    # Web URL would look like http://<server>/<component>.<ext> 
    # path = <component>.<ext> identifies what is being requested 
    # base of the path is the component name
    # the extension indicates what output gets returned

    # no component at all gives the list of known components
    # component but no extension gives the main design window

    ext = ""
    if '.' in path:
        try:
            path, ext = path.split('.')
        except ValueError as e:
            print("*** ERROR ***", repr(e))
            traceback.print_exc()
            abort(400)

    component = path

    if component:
        # Specific component selected
        try:
            # Get the python object according to the component name
            f = getComponent(component)
        except Exception as e:
            # Can't find component in library
            print("*** ERROR ***", repr(e))
            traceback.print_exc()
            abort(404)

        if ext == "json":
            # Returns component parameters and their metadata
            return jsonify(f.getParameterInfo())
        elif ext == "ajax":
            # Returns all outputs suitable for use by a frontend interface
            ss = make(f, component, request.form)
            if "png" in ss:
                ss["png"] = str(base64.b64encode(ss["png"]), "utf-8")
            return jsonify(ss)
        elif ext in known_types:
            # Returns a single output for download by the user
            ss = make(f, component, request.args)
            return Response(ss[known_types[ext][0]], mimetype=known_types[ext][1])
        elif ext: 
            # Unknown extension
            abort(400)

        else:
            # Create the main design interface
            # Set makeOutput arguments appropriately
            make(f, component, request.args)

            # Set specific parameters if provided, allows hotlinking to specific design instances
            t = max(known_models.get(component, {}).get("t", 0), MIN_THICKNESS)
            paperthick = (t == MIN_THICKNESS and "selected" or "")
            woodthick = (t == 3 and "selected" or "")

            # Create a form dynamically based on the parameters in the selected component
            class ParamForm(Form):
                pass
            for k, v in f.parameters.items():
                # Add an input field to the form based on the parameter type
                # Skip parameters that start with '_'
                if k[0] != '_' and '._' not in k:
                    setattr(ParamForm, k, MyField("%s: " % k, v.getSpec(), default=v.getValue()))

            # Render the main design interface
            component_paths = getComponentPaths(component)
            if session.get("basic", False):
                html = "basic.html"
            else:
                html = "design_component.html"

            return render_template(html,
                component=component, 
                paperthick=paperthick, 
                woodthick=woodthick, 
                query=request.query_string, 
                form=ParamForm(), git_files=git_files, source_code_paths=component_paths)

# Debug Mode set by env variable FLASK_DEBUG
if __name__ == "__main__":
    debug = os.environ.get("FLASK_DEBUG", False)
    app.run(debug=debug,
            host='0.0.0.0',
            port=5000,
            extra_files=['./static/css/component_ui.css'])
